#if defined( CYPRESS )
#include <m8c.h>        // part specific constants and macros
#include "delay.h"
#else
#include "stdio.h"
#endif
#include "stdlib.h"
#include "ulisp.h"
#include "ulisp_defs.h"

#if defined( WIN32 )
#include "windows.h"
#endif

#if defined( WIN32 )
#define delay Sleep
#elif defined( CYPRESS ) 
void delay(int d) {
	do {
		if (d > 240) {
#if 1
			// for 24MHz speed:
			delay_ms(240);
#else
			// for 3MHz default:
			delay_ms(30);
#endif
			d -= 240;
		} else {
			delay_ms(d);
			d = 0;
		}
	} while (d);
}
#else
#error
#endif

#if defined( WIN32 )
char program[1024]; 
#elif defined( CYPRESS ) 
#include "relay.h"
#define program program.program
#else
#error
#endif

int lisp_RET, lisp_PC = 0, lisp_I, lisp_J, lisp_K;
#if !defined( CYPRESS )
int lisp_A, lisp_B, lisp_C, lisp_D, lisp_X, lisp_Y, lisp_Z;
#endif

void lisp_skip(void) {
	char sp = 1;
	while (sp) {
		if (program[lisp_PC] == SYM_START) {
			sp++;
		}
		if (program[lisp_PC] == SYM_END) {
			sp--;
		}
		lisp_PC++;
	}
}

#define GET_PORT(a,b) (((a)&(1<<b))?1:0)
#define SET_PORT(a,b) ((a)|=(1<<b))
#define CLR_PORT(a,b) ((a)&=~(1<<b))

int lisp_get(char par) {
	switch (par) {
	case SYM_A: 
#if defined( CYPRESS )
		return GET_PORT(P0DATA, 6);
#else
		return lisp_A;
#endif
	case SYM_B: 
#if defined( CYPRESS )
		return GET_PORT(P0DATA, 5);
#else
		return lisp_B;
#endif
	case SYM_C: 
#if defined( CYPRESS )
		return GET_PORT(P0DATA, 4);
#else
		return lisp_C;
#endif
	case SYM_D: 
#if defined( CYPRESS )
		return GET_PORT(P0DATA, 3);
#else
		return lisp_D;
#endif
	case SYM_X: 
#if defined( CYPRESS )
		return GET_PORT(P1DATA, 6);
#else
		return lisp_X;
#endif
	case SYM_Y: 
#if defined( CYPRESS )
		return GET_PORT(P1DATA, 5);
#else
		return lisp_Y;
#endif
	case SYM_Z: 
#if defined( CYPRESS )
		return GET_PORT(P1DATA, 4);
#else
		return lisp_Z;
#endif
	case SYM_I: 
		return lisp_I;
	case SYM_J: 
		return lisp_J;
	case SYM_K: 
		return lisp_K;
	case SYM_0: 
		return 0;
	case SYM_1: 
		return 1;
	case SYM_2: 
		return 2;
	case SYM_3: 
		return 3;
	case SYM_4: 
		return 4;
	case SYM_5: 
		return 5;
	case SYM_6: 
		return 6;
	case SYM_7: 
		return 7;
	case SYM_8: 
		return 8;
	case SYM_9: 
		return 9;
	default:
#if defined( DEBUG ) && defined( WIN32 )
		printf("lisp_get(), unk %d \'%c\'\r\n", par, par);
#endif
		return -1;
	}
}

void lisp_set(char par, int val) {
	switch (par) {
	case SYM_A: 
#if defined( CYPRESS )
		if (val&1) {
			SET_PORT(P0DATA, 6);
		} else {
			CLR_PORT(P0DATA, 6);
		}
#else
		lisp_A = val; 
#endif
		break;
	case SYM_B: 
#if defined( CYPRESS )
		if (val&1) {
			SET_PORT(P0DATA, 5);
		} else {
			CLR_PORT(P0DATA, 5);
		}
#else
		lisp_B = val; 
#endif
		break;
	case SYM_C: 
#if defined( CYPRESS )
		if (val&1) {
			SET_PORT(P0DATA, 4);
		} else {
			CLR_PORT(P0DATA, 4);
		}
#else
		lisp_C = val; 
#endif
		break;
	case SYM_D: 
#if defined( CYPRESS )
		if (val&1) {
			SET_PORT(P0DATA, 3);
		} else {
			CLR_PORT(P0DATA, 3);
		}
#else
		lisp_D = val; 
#endif
		break;
	case SYM_X: 
#if defined( CYPRESS )
		if (val&1) {
			SET_PORT(P1DATA, 6);
		} else {
			CLR_PORT(P1DATA, 6);
		}
#else
		lisp_X = val; 
#endif
		break;
	case SYM_Y: 
#if defined( CYPRESS )
		if (val&1) {
			SET_PORT(P1DATA, 5);
		} else {
			CLR_PORT(P1DATA, 5);
		}
#else
		lisp_Y = val; 
#endif
		break;
	case SYM_Z: 
#if defined( CYPRESS )
		if (val&1) {
			SET_PORT(P1DATA, 4);
		} else {
			CLR_PORT(P1DATA, 4);
		}
#else
		lisp_Z = val; 
#endif
		break;
	case SYM_I: 
		lisp_I = val; 
		break;
	case SYM_J: 
		lisp_J = val; 
		break;
	case SYM_K: 
		lisp_K = val; 
		break;
	default:
#if defined( WIN32 ) && defined( DEBUG )
		printf("lisp_set(), unk %d \'%c\'\r\n", par, par);
#endif
		break;
	}
}

int lisp_proc(void) {
	char cmd = 0, var = 0, par = 0;
	int ret = 0;
	int sp = lisp_PC-1;
read:
	var = program[lisp_PC];
	lisp_PC++;
	if (var == SYM_EOF) {
		lisp_PC = 0;
		goto read;
	} else
	if (var == SYM_START) {
		if (cmd == SYM_WHILE || cmd == SYM_IF) {
			if (par == 1 || par == 0) {
				lisp_RET = lisp_proc();
			} else {
				lisp_skip();
			}
		} else {
			lisp_RET = lisp_proc();
		}
		if (cmd) {
			goto work;
		}
		goto read;
	} else 
	if (var == SYM_END) {
		if (cmd == SYM_WHILE && par == 1) {
			lisp_PC = sp;
		}
#if defined( WIN32 )
		if (cmd == SYM_PRINT) {
			printf("\r\n");
		}
#endif
		return ret;
	} else
	if (var == SYM_STOP) {
#if defined( WIN32 )
		exit(0);
#elif defined( CYPRESS )
		while (1);
#else
#error
#endif
	} else {
		if (cmd == 0) {
			cmd = var;
#if defined( DEBUG ) && defined( WIN32 )
			if (cmd == SYM_PRINT) {
				printf("PRINT: ");
				//printf("STATE: %d%d%d%d %d%d%d pc=%d\r\n",
				//	lisp_get('A'), lisp_get('B'), lisp_get('C'), lisp_get('D'), 
				//	lisp_get('X'), lisp_get('Y'), lisp_get('Z'),
				//	lisp_PC);
			}
#endif
		} else {
			if (var == SYM_CONST1) {
				lisp_RET = (unsigned char)program[lisp_PC];
				lisp_PC++;
			} else
			if (var == SYM_CONST2) {
				lisp_RET  = (unsigned char)program[lisp_PC];
				lisp_RET += (unsigned char)program[lisp_PC+1]<<8;
				lisp_PC  += 2;
#if defined( WIN32 )
			} else
			if (var == SYM_CONST3) {
				lisp_RET  = program[lisp_PC];
				lisp_RET += program[lisp_PC+1]<<8;
				lisp_RET += program[lisp_PC+2]<<16;
				lisp_PC  += 3;
			} else
			if (var == SYM_CONST4) {
				lisp_RET  = program[lisp_PC];
				lisp_RET += program[lisp_PC+1]<<8;
				lisp_RET += program[lisp_PC+2]<<16;
				lisp_RET += program[lisp_PC+3]<<24;
				lisp_PC  += 4;
#endif
			} else {
				if (par == 0) { 
					if (cmd != SYM_SETQ) {
						lisp_RET = lisp_get(var);
					}
				} else {
					lisp_RET = lisp_get(var);
				}
			}
work:
			if (cmd == SYM_WHILE || cmd == SYM_IF) {
				if (par == 0) {
					if (lisp_RET) {
						par = 1;
					} else {
						par = 2;
					}
				}
			} else
			if (cmd == SYM_SETQ) {
				if (par == 0) {
					par = var;
					goto read;
				}
			} else {
				par++;
			}
			// par:
			// set -  ,
			// while,if - 0 - , 1 - , 2 -  ,
			//  -  .
			switch (cmd) {
			case SYM_ADD:
				ret += lisp_RET;
				break;
			case SYM_SUB:
				if (par == 1) {
					ret = lisp_RET;
				} else {
					ret -= lisp_RET;
				}
				break;
			case SYM_AND:
				if (par == 1) {
					ret = lisp_RET;
				} else {
					ret &= lisp_RET;
				}
				break;
			case SYM_DIV:
				if (par == 1) {
					ret = lisp_RET;
				} else {
					ret /= lisp_RET;
				}
				break;
			case SYM_MUL:
				if (par == 1) {
					ret = lisp_RET;
				} else {
					ret *= lisp_RET;
				}
				break;
			case SYM_OR:
				ret |= lisp_RET;
				break;
			case SYM_XOR:
				ret ^= lisp_RET;
				break;
			case SYM_EQ:
				if (par == 1) {
					ret = lisp_RET;
				} else {
					ret = (ret == lisp_RET) ? 1 : 0;
				}
				break;
			case SYM_DELAY:
				ret = lisp_RET;
				delay(lisp_RET);
				break;
			case SYM_MORE:
				if (par == 1) {
					ret = lisp_RET;
				} else {
					ret = (ret > lisp_RET) ? 1 : 0;
				}
				break;
			case SYM_LESS:
				if (par == 1) {
					ret = lisp_RET;
				} else {
					ret = (ret < lisp_RET) ? 1 : 0;
				}
				break;
			case SYM_SETQ:
				lisp_set(par, lisp_RET);
				ret = lisp_RET;
				break;
			case SYM_PRINT:
#if defined( DEBUG ) && defined( WIN32 )
				if (var == SYM_CONST1 || var == SYM_CONST2 || var == SYM_CONST3 || var == SYM_CONST4) {
					printf("\'%c\' = %d, ", 'c', lisp_RET);
				} else {
					printf("\'%c\' = %d, ", var, lisp_RET);
				}
#endif
				break;
			default:
				ret = lisp_RET;
			}
		}
		goto read;
	}
	return ret;
}

#if defined( WIN32 ) && defined( DEBUG )
void main(int argc, char **argv) {
	//     :
	memset(program, 0x00, sizeof(program));
	//    :
	if (argc == 2) {
		FILE *f;
		int size = 0;
		f = fopen(argv[1], "r");
		if (f == NULL) {
			printf("error: bad input file \'%s\'!\r\n", argv[1]);
			exit(1);
		}
		while (1) {
			int v = fgetc(f);
			if (v == EOF) {
				break;
			} else {
				program[size] = v;
				size++;
			}
		}
		fclose (f);
	} else {
		printf("error: no input file!");
		exit(1);
	}
	// :
	lisp_PC = 1;
	while (1) {
		lisp_proc();
	}
}
#endif
