#include "stdlib.h"
#include "stdio.h"
#include "ulisp_defs.h"

#include "windows.h"

char raw_program[512*128];
char pre_program[512*128];
int size;

char program[512];
int lisp_PC = 0;
int lisp_OFFSET = 0;

#define MAX_STACK_DEPTH 10

void lisp_put_pre(char v) {
	if (lisp_PC < sizeof(program)) {
		pre_program[lisp_PC] = v;
		lisp_PC++;
	} else {
		printf("error: too big program!\r\n");
		exit(1);
	}
}

void lisp_put(char v) {
	if (lisp_PC < sizeof(program)) {
		program[lisp_PC] = v;
		lisp_PC++;
	} else {
		printf("error: too big program!\r\n");
		exit(1);
	}
}

void lisp_pre(void) {
	int i = 0;
	int offset = 0;
	int comment = 0;
	int space = 1;
	int line = 0;
	while (i < size) {
		char v = raw_program[i];
		i++;
		if (v == '\n') {
			if (space == 0) {
				lisp_put_pre(' ');
			}
			comment = 0;
			space = 1;
			line++;
			continue;
		} else 
		if (comment == 1) {
			continue;
		} else
		if (v == ';') {
			comment = 1;
			continue;
		} else
		if (v == '\t') {
			if (space == 0) {
				lisp_put_pre(' ');
			}
			space = 1;
			continue;
		} else
		if (v == '(' || v == ')') {
			lisp_put_pre(v);
			space = 1;
			continue;
		} else
		if (v == ' ') {
			if (space == 0) {
				lisp_put_pre(v);
			}
			space = 1;
			continue;
		} else
		if (v == '\r') {
			// nop
			continue;
		}
		space = 0;
		lisp_put_pre(v);
	}
}

void lisp_check_stack(void) {
	char v;
	int stack = 0;
	int offset = 0;
	int max_depth = 0;
	while (offset < sizeof(pre_program)) {
		v = pre_program[offset];
		offset++;
		if (v == SYM_START) {
			stack++;
		}
		if (max_depth < stack) {
			max_depth = stack;
		}
		if (stack > MAX_STACK_DEPTH) {
			printf("error: stack overflow!\r\n");
			exit(1);
		}
		if (v == SYM_END) {
			stack--;
		}
		if (stack > MAX_STACK_DEPTH) {
			printf("error: stack error!\r\n");
			exit(1);
		}
	}
	if (stack != 0) {
		printf("error: final stack error!\r\n");
		exit(1);
	}
	printf("info: stack checked (max depth %d of %d)!\r\n", max_depth, MAX_STACK_DEPTH);
}

typedef struct {
	char name[30];
	char sym;
} command_t, *pcommand_t;

command_t commands[] = {
	{ "PRINT",  'p' },
	{ "IF",     'i' },
	{ "WHILE",  'w' },
	{ "DELAY",  'd' },
	{ "STOP",   'x' },
	{ "SETQ",   's' },
	{ ">",      '>' },
	{ "MORE",   '>' },
	{ "=",      '=' },
	{ "EQ",     '=' },
	{ "<",      '<' },
	{ "LESS",   '<' },
	{ "/=",     '#' },
	{ "!=",     '#' },
	{ "NEQ",    '#' },
	{ "+",      '+' },
	{ "ADD",    '+' },
	{ "-",      '-' },
	{ "SUB",    '-' },
	{ "*",      '*' },
	{ "MUL",    '*' },
	{ "/",      '/' },
	{ "DIV",    '/' },
	{ "&",      '&' },
	{ "AND",    '&' },
	{ "|",      '|' },
	{ "OR",     '|' },
	{ "^",      '^' },
	{ "XOR",    '^' },
};

lisp_command() {
	char v;
	int s = 0;
	int i;
	while (1) {
		v = pre_program[lisp_OFFSET+s];
		if (v == ' ' || v == SYM_END) {
			break;
		}
		s++;
	}
	for (i = 0; i < sizeof(commands)/sizeof(command_t); i++) {
		if (strlen(commands[i].name) == s) {
			if (memcmp(commands[i].name, &pre_program[lisp_OFFSET], s) == 0) {
				lisp_put(commands[i].sym);
				break;
			}
		}
	}
	if (i == sizeof(commands)/sizeof(command_t)) {
		pre_program[lisp_OFFSET+s] = 0;
		printf("error: unknown command \'%s\'!\r\n", &pre_program[lisp_OFFSET]);
		exit(1);
	}
	lisp_OFFSET += s;
}

lisp_param() {
	char v;
	int s = 0;
	int i;
	while (1) {
		v = pre_program[lisp_OFFSET+s];
		if (v == ' ' || v == SYM_END) {
			break;
		}
		s++;
	}
	if (s == 1) {
		switch (pre_program[lisp_OFFSET]) {
		case SYM_A:
			lisp_put('A');
			break;
		case SYM_B:
			lisp_put('B');
			break;
		case SYM_C:
			lisp_put('C');
			break;
		case SYM_D:
			lisp_put('D');
			break;
		case SYM_X:
			lisp_put('X');
			break;
		case SYM_Y:
			lisp_put('Y');
			break;
		case SYM_Z:
			lisp_put('Z');
			break;
		case SYM_I:
			lisp_put('I');
			break;
		case SYM_J:
			lisp_put('J');
			break;
		case SYM_K:
			lisp_put('K');
			break;
		case SYM_0:
			lisp_put('0');
			break;
		case SYM_1:
			lisp_put('1');
			break;
		case SYM_2:
			lisp_put('2');
			break;
		case SYM_3:
			lisp_put('3');
			break;
		case SYM_4:
			lisp_put('4');
			break;
		case SYM_5:
			lisp_put('5');
			break;
		case SYM_6:
			lisp_put('6');
			break;
		case SYM_7:
			lisp_put('7');
			break;
		case SYM_8:
			lisp_put('8');
			break;
		case SYM_9:
			lisp_put('9');
			break;
		default:
			printf("error: bad parameter \'%c\'\r\n", pre_program[lisp_OFFSET]);
		}
	} else {
		int value = 0;
		int sign = 0;
		if (pre_program[lisp_OFFSET] == '-') {
			sign = 1;
			i = 1;
		} else {
			i = 0;
		}
		for (i = 0; i < s; i++) {
			if (pre_program[lisp_OFFSET+i] <= '9' &&
				pre_program[lisp_OFFSET+i] >= '0') {
				value *= 10;
				value += pre_program[lisp_OFFSET+i] - '0';
			} else {
				pre_program[lisp_OFFSET+s] = 0;
				printf("error: bad parametrer \'%s\'\r\n", pre_program[lisp_OFFSET]);
			}
		}
		if (value < (1<<7)) {
			lisp_put(SYM_CONST1);
			if (sign) {
				lisp_put((unsigned char)-value);
			} else {
				lisp_put((unsigned char)value);
			}
		} else if (value < (1<<15)) {
			lisp_put(SYM_CONST2);
			if (sign) {
				lisp_put(((unsigned short)-value)&0xFF);
				lisp_put(((unsigned short)-value)>>8);
			} else {
				lisp_put((unsigned short)value&0xFF);
				lisp_put((unsigned short)value>>8);
			}
		} else {
			printf("error: constant \'%d\' out of bound!\r\n", value);
		} 
	}
	lisp_OFFSET += s;
}

int main(int argc, char **argv) {
	FILE *f;
	int par;

	size = 0;
	
	if (argc == 1) {
		printf("help: lisp_compile.exe input.list [output.ls]\r\n");
		exit(1);
	}
	
	f = fopen(argv[1],"r");
	if (f == NULL) {
		printf("error: bad input file \'%s\'!\r\n", argv[1]);
		exit(1);
	}
	while (1) {
		int v = fgetc(f);
		if (v == EOF) {
			break;
		} else {
			raw_program[size] = v;
			size++;
		}
	}
	fclose (f);

	printf("info: loaded %d bytes!\r\n", size);
	
	lisp_PC = 0;
	lisp_pre();
	lisp_check_stack();
	
	size = lisp_PC;
	lisp_PC = 0;
	
	printf("info: after preprocessor size %d bytes!\r\n", size);
	
	while (lisp_OFFSET<size) {
		char v;
		v = pre_program[lisp_OFFSET];
		if (v == SYM_START) {
			lisp_put(v);
			par = 0;
			lisp_OFFSET++;
			continue;
		} else
		if (v == SYM_END) {
			lisp_put(v);
			lisp_OFFSET++;
			continue;
		} else
		if (v == ' ') {
			par++;
			lisp_OFFSET++;
			continue;
		}
		if (par == 0) {
			lisp_command();
		} else {
			lisp_param();
		}
	}
	
	if (argc > 2) {
		f = fopen(argv[2],"wb");
		if (f == NULL) {
			printf("error: bad output file \'%s\'!\r\n", argv[2]);
			exit(1);
		}
		fwrite(program, 1, lisp_PC, f);
		fclose (f);
	} else {
		int i;
		printf("char program[] = {");
		for (i = 0; i<lisp_PC; i++) {
			if (program[i] == '(') printf("\'%c\',", program[i]); else
			if (program[i] == ')') printf("\'%c\',", program[i]); else
			if (program[i] == '=') printf("\'%c\',", program[i]); else
			if (program[i] == '#') printf("\'%c\',", program[i]); else
			if (program[i] == '+') printf("\'%c\',", program[i]); else
			if (program[i] == '-') printf("\'%c\',", program[i]); else
			if (program[i] == '*') printf("\'%c\',", program[i]); else
			if (program[i] == '/') printf("\'%c\',", program[i]); else
			if (program[i] == '&') printf("\'%c\',", program[i]); else
			if (program[i] == '|') printf("\'%c\',", program[i]); else
			if (program[i] == '^') printf("\'%c\',", program[i]); else
			if (program[i] == '!') printf("\'%c\',", program[i]); else
			if (program[i] == '@') printf("\'%c\',", program[i]); else
			if (program[i] == '!') printf("\'%c\',", program[i]); else
			if (program[i] == '[') printf("\'%c\',", program[i]); else
			if (program[i] == ']') printf("\'%c\',", program[i]); else
			if (program[i] == ';') printf("\'%c\',", program[i]); else			
			if (program[i] == ':') printf("\'%c\',", program[i]); else			
			if (program[i] == '\'') printf("\'%c\',", program[i]); else			
			if (program[i] == '\"') printf("\'%c\',", program[i]); else			
			if (program[i] == '<') printf("\'%c\',", program[i]); else			
			if (program[i] == '>') printf("\'%c\',", program[i]); else			
			if (program[i] == '?') printf("\'%c\',", program[i]); else			
			if (program[i] == '/') printf("\'%c\',", program[i]); else			
			if (program[i] == '%') printf("\'%c\',", program[i]); else			
			if (program[i] <= 'Z' && program[i] >= 'A') {
				printf("\'%c\',", program[i]);
			} else
			if (program[i] <= 'z' && program[i] >= 'a') {
				printf("\'%c\',", program[i]);
			} else
			if (program[i] <= '9' && program[i] >= '0') {
				printf("\'%c\',", program[i]);
			} else {
				printf("0x%02x,", (unsigned char)program[i]);
			}
		}
		printf("};\r\n");
	}
	
	return 0;
}
